#include <avr/wdt.h> // подключаем библиотеку для работы со сторожевой собакой, или как там оно по-русски
#include <OneWire.h> // подключаем библиотеку для работы с шиной 1-Wire

#define MPX_ADC          A1     // на эту ногу подключим датчик давления
#define MPX_ADC_SAMPLES  16.0   // выводим среднее из MPX_ADC_SAMPLES последовательных измеренеий
#define MPX_ADC_ZERO     512.0  // показания АЦП при одинаковом давлении на обоих портах

#define W1_BUS           A3     // эту ножку процессора будем использовать для работы с шиной
#define REPORT_INTERVAL  1000   // отчеты в ПК будем послыать раз в секунду (1000 мс)

//#define COMPLETE_DELTAS         // для каждой пары градусников вывести дельту дважды, в виде положительного и отрицательного значения
#ifndef COMPLETE_DELTAS         // в противном случае, возможно, захочется
//  #define ABSOLUTE_DELTAS       // выводить абсолютное значение дельты, без знака
#endif

#define MAX_SENSORS      8      // лучше написать тут реальное количество градусников

#define SERIAL_MATCH(a,b)  (a[0]==b[0]&&a[1]==b[1]&&a[2]==b[2]&&a[3]==b[3]&&a[4]==b[4]&&a[5]==b[5])

typedef struct {
  uint8_t serial[6]; // 6
  float value;       // 4
} ds18b20_t;  //  датчик

typedef struct {
  uint8_t serial[6]; // 6
  float a;           // 4
  float b;           // 4
  float c;           // 4
} ds18b20adj_t;  //  поправка

ds18b20adj_t adj[MAX_SENSORS];   // массив поправок
uint8_t adjc = 0;

void setup()
{
  // поправки   ( шесть байт серийный номер датчика, a,   b,   c   )
  add_adjustment( 0xf2, 0x36, 0x24, 0x03, 0x0, 0x0,  0.0, 1.0, 2.0 );  // серийный номер 0xF23624030000
  add_adjustment( 0x91, 0x02, 0x24, 0x03, 0x0, 0x0,  0.0, 1.0, 3.0 );  // серийный номер 0x910224030000
  add_adjustment( 0x44, 0xb0, 0x09, 0x06, 0x0, 0x0,  0.0, 2.0, 0.0 );  // серийный номер 0x44B009060000
//  add_adjustment( 0x0, 0x0, 0x0, 0x0, 0x0, 0x0,  0.0, 1.0, 0.0 );
//  ...
  
  // инициализируем ADC
  pinMode(MPX_ADC, INPUT);
  //analogReference(EXTERNAL);
  
  Serial.begin(115200);  // 115200/8/N/1
  wdt_enable(WDTO_1S);
}

void loop()
{
  static OneWire ds(W1_BUS);
  static unsigned long tr = 0, tc = 0;
  
  byte i, j, cnt, cfg, data[5], addr[8];
  int16_t raw;
  unsigned long t = millis();
  ds18b20_t sens[MAX_SENSORS];
  float d, adc, volt;
  
  if (tc <= t) // посылаем всем DS18B20 команду приступить к измерениям
  {
    tc += REPORT_INTERVAL; // сохраняем время начала следующего цикла измерений

    cnt = 0;  // обнуляем счетчик датчиков
    ds.reset_search();
    while (ds.search(addr)) // сканируем шины
    {
      if (OneWire::crc8(addr, 7) != addr[7]) continue; // ошибка контрольной суммы
      if (addr[0] != 0x28) continue; // не ds18b20
      
      ds.reset();
      ds.select(addr);
      ds.write(0x44); // собственно команда приступить к измерениям

      if (++cnt >= MAX_SENSORS) break;  // ограничение количества датчиков
    }

    tr = t + 800; // сохраняем время начала чтения измерений
  }

  if (tr && tr <= t) // читаем измерения со всех подключенных датчиков
  {
    tr = 0;

    cnt = 0;  // обнуляем счетчик датчиков
    ds.reset_search();
    while (ds.search(addr))  // сканируем шину
    {
      if (OneWire::crc8(addr, 7) != addr[7]) continue; // кривой CRC
      if (addr[0] != 0x28) continue; // не ds18b20
      
      ds.reset();
      ds.select(addr);    
      ds.write(0xbe);  // собственно команда чтения
      
      for (i = 0; i < 5; i++) data[i] = ds.read(); // читаем пять байт
      ds.reset();
      raw = (data[1] << 8) | data[0]; // из первых двух собираем результат измерений

      cfg = (data[4] & 0x60); // из пятого байта определяем разрешение датчика, соответственно ровняем цифру
      if (cfg == 0x00) raw = raw&~7;      // 9 бит
      else if (cfg == 0x20) raw = raw&~3; // 10 бит
      else if (cfg == 0x40) raw = raw&~1; // 11 бит
  
      for (i = 1; i < 7; i++) sens[cnt].serial[i-1] = addr[i]; // сохраняем серийный номер 
      sens[cnt].value = (float)raw / 16.0;  // сохраняем показания 
      
      if (++cnt >= MAX_SENSORS) break;  // ограничение количества датчиков
    }
    
    // cчитаем поправки
    for (i = 0; i < cnt; i++) // перебираем датчики
    {
      for (j = 0; j < adjc; j++) // перебираем поправки
      {
        if (SERIAL_MATCH(sens[i].serial, adj[j].serial)) // ищем по совпадению серийного номера
        {
          sens[i].value = adj[j].a * sens[i].value * sens[i].value + adj[j].b * sens[i].value + adj[j].c;  // T' = A*T^2 + B*T + C
          break; // один градусник - одна поправка
        }
      }
    }
    // ---
    
    // меряем давление
    adc = 0;
    for (i = 0; i < MPX_ADC_SAMPLES; i++) // читаем аналог MPX_ADC_SAMPLES раз
    {
      adc += analogRead(MPX_ADC); // суммируем
      delay(1);
    }
    adc /= MPX_ADC_SAMPLES; // усредняем
    
    // выведем реальное значение серединки диапазона при одинаковом давлении на обоих портах, 
    // потом напишем его в MPX_ADC_ZERO на шестой строчке, и закоментируем следующую строчку
    Serial.println(adc);
    
    adc -= MPX_ADC_ZERO;
    volt = (5.0 * adc) / 1024.0; // 1V=1KPa
    // ---
    
    
    // выводим отчет
    Serial.println('{');  // токен начала отчета
    
    for (i = 0; i < cnt; i++) // вывод показаний
    {
      print_serial(sens[i].serial);  // серийный номер 
      Serial.print('<');             // тип устройства
      Serial.println(sens[i].value); // показания

#ifdef COMPLETE_DELTAS 
      for (j = 0; j < cnt; j++) // цикл вывода положительного и отрицательного значения дельты
      {
        if (i == j) continue;
#else
      for (j = i + 1; j < cnt; j++)  // короткий цикл вывода со знаком как получится
      {
#endif
        d = sens[i].value - sens[j].value; // считаем дельту

#ifdef ABSOLUTE_DELTAS
        if (d < 0) d = -d;
#endif
        print_serial(sens[i].serial);  // выводим серийник первого датчика
        Serial.print('-');             // отделяем его от второго
        print_serial(sens[j].serial);  // выводим серийник второго датчика
        Serial.print('<');             // тип устройства
        Serial.println(d);             // дельта
      }
    }

    Serial.print('P'); // давление
    Serial.print('<');
    Serial.println(volt * 1000.0); // Па

    Serial.println('}');  // токен конца отчета
  }
  
  wdt_reset(); // сбрасываем счетчик сторожевого собачка
}


void print_serial (uint8_t *s) // вывод серийного номера датчика
{
  for (uint8_t i = 0; i < 6; i++)
  {
     if (s[i] < 0x10) Serial.print(0);
     Serial.print(s[i], HEX);
  }
}

// удобно конфигурируем поправки и отслеживаем их количество
void add_adjustment (int8_t s0, int8_t s1, int8_t s2, int8_t s3, int8_t s4, int8_t s5, float a, float b, float c)
{
  if (adjc >= MAX_SENSORS) return;
  adj[adjc].serial[0] = s0;
  adj[adjc].serial[1] = s1;
  adj[adjc].serial[2] = s2;
  adj[adjc].serial[3] = s3;
  adj[adjc].serial[4] = s4;
  adj[adjc].serial[5] = s5;
  adj[adjc].a = a;
  adj[adjc].b = b;
  adj[adjc++].c = c;
}
